@@ -14,8 +14,8 @@ from configparser import ConfigParser
 from minitrino import utils
 from minitrino import errors as err
 from minitrino.settings import RESOURCE_LABEL
-from minitrino.settings import MODULE_LABEL_KEY_ROOT
 from minitrino.settings import MODULE_ROOT
+from minitrino.settings import MODULE_ADMIN
 from minitrino.settings import MODULE_SECURITY
 from minitrino.settings import MODULE_CATALOG
 
@@ -49,7 +49,6 @@ class Environment:
 
     @utils.exception_handler
     def __init__(self):
-
         # Attributes that depend on user input prior to being set
         self.verbose = False
         self._user_env = []
@@ -85,7 +84,7 @@ class Environment:
         try:
             # Try to get `LIB_path` var - handle exception if `env` attribute is
             # not yet set
-            lib_dir = self.env.get_var("LIB_PATH", "")
+            lib_dir = self.env.get("LIB_PATH", "")
         except:
             pass
 
@@ -129,9 +128,8 @@ class Environment:
         # Skip the library-related procedures if the library is not found
         try:
             if self.minitrino_lib_dir:
-                self.logger.log(
+                self.logger.verbose(
                     f"Library path set to: {self.minitrino_lib_dir}",
-                    level=self.logger.verbose,
                 )
 
             # Now that we know where the library is, we can try to parse the env
@@ -143,12 +141,11 @@ class Environment:
             cli_ver = utils.get_cli_ver()
             lib_ver = utils.get_lib_ver(self.minitrino_lib_dir)
             if cli_ver != lib_ver:
-                self.logger.log(
+                self.logger.warn(
                     f"CLI version {cli_ver} and library version {lib_ver} "
                     f"do not match. You can update the Minitrino library "
                     f"version to match the CLI version by running 'minitrino "
-                    f"lib_install'.",
-                    level=self.logger.warn,
+                    f"lib-install'.",
                 )
         except:
             pass
@@ -176,118 +173,93 @@ class Environment:
 
         config_file = os.path.join(self.minitrino_user_dir, "minitrino.cfg")
         if not os.path.isfile(config_file):
-            self.logger.log(
+            self.logger.warn(
                 f"No minitrino.cfg file found at {config_file}. "
                 f"Run 'minitrino config' to reconfigure this file and directory.",
-                level=self.logger.warn,
             )
         return config_file
 
     def _get_docker_clients(self):
-        """Gets DockerClient and APIClient objects. References the DOCKER_HOST
-        variable in `minitrino.cfg` and uses for clients if present. Returns a
-        tuple of DockerClient and APIClient objects, respectiveley.
+        """Gets DockerClient and APIClient objects. Returns a tuple of DockerClient
+        and APIClient objects, respectively.
 
         If there is an error fetching the clients, None types will be returned
         for each client. The lack of clients should be caught by check_daemon()
         calls that execute in each command that requires an accessible Docker
         service."""
 
+        docker_url = os.environ.get("DOCKER_HOST", "")
+
         try:
-            docker_host = self.env.get_var("DOCKER_HOST", "")
-            docker_client = docker.DockerClient(base_url=docker_host)
-            api_client = docker.APIClient(base_url=docker_host)
+            docker_client = docker.DockerClient(base_url=docker_url)
+            api_client = docker.APIClient(base_url=docker_url)
             self.docker_client, self.api_client = docker_client, api_client
         except:
             return None, None
 
 
-class EnvironmentVariables:
-    """Gathers all Minitrino variables into a single source of truth.
+class EnvironmentVariables(dict):
+    """Exposes all Minitrino variables.
 
     ### Parameters
     - `ctx`: Instantiated Environment object (with user input already accounted
       for).
 
-    ### Public Attributes
-    - `env`: Dictionary containing all environment variables.
-
-    ### Public Methods
-    - `get_var()`: Gets an environment variable from a specific section and key.
-    - `get_section()`: Gets a a section from the environment variable dict.
-
     ### Usage
     ```python
     # ctx object has an instantiated EnvironmentVariables object
-    env_variable = ctx.env.get_var("STARBURST_VER", "354-e")
-    env_section = ctx.env.get_section("MODULES")
+    env_variable = ctx.env.get("STARBURST_VER", "388-e")
     ```"""
 
     @utils.exception_handler
     def __init__(self, ctx=None):
-
+        super().__init__()
         if not ctx:
             raise utils.handle_missing_param(list(locals().keys()))
 
-        self.env = {}
         self._ctx = ctx
 
-        self._parse_minitrino_config()
         self._parse_user_env()
+        self._parse_os_env()
+        self._parse_minitrino_config()
 
-    def get_var(self, key="", default=None):
-        """Gets and returns a variable from a section of the environment
-        variables. Since it is assumed there will not be duplicate environment
-        variables between sections, the first-match is returned.
+    def _parse_user_env(self):
+        """Parses user-provided environment variables for the current
+        command. Highest precedence in environment variable order, loaded
+        first."""
 
-        ### Parameters
-        - `key`: The key to search for.
-        - `default` The default value to return if the key is not found."""
+        if not self._ctx._user_env:
+            return
 
-        if not key:
-            raise utils.handle_missing_param(["key"])
+        for env_var in self._ctx._user_env:
+            env_var = utils.parse_key_value_pair(env_var, err_type=err.UserError)
+            self[env_var[0]] = env_var[1]
 
-        for section_k, section_v in self.env.items():
-            if isinstance(section_v, dict):
-                for section_dict_k, section_dict_v in section_v.items():
-                    if section_dict_k.upper() == key.upper():
-                        return section_dict_v
-        return default
+    def _parse_os_env(self):
+        """Parses environment variables from the user's shell. Middle
+        precedence in environment variable order, loaded second."""
 
-    def get_section(self, section=""):
-        """Gets and returns a section from the environment variables. If the
-        section doesn't exist, an empty dict is returned.
-
-        ### Parameters
-        - `section`: The section to return, if it exists."""
-
-        if not section:
-            raise utils.handle_missing_param(list(locals().keys()))
-
-        return self.env.get(section.upper(), {})
+        append = ["LIB_PATH", "STARBURST_VER", "TEXT_EDITOR", "LIC_PATH"]
+        for k, v in os.environ.items():
+            k = k.upper()
+            if k in append and not self.get(k, None):
+                self[k] = v
 
     def _parse_minitrino_config(self):
         """Parses the Minitrino config file and adds it to the env
-        dictionary."""
+        dictionary. Middle precedence in environment variable order, loaded
+        third."""
 
         if not os.path.isfile(self._ctx.config_file):
             return
 
-        # Catch this and provide a useful message, as it can be tricky to track
-        # down otherwise
         try:
             config = ConfigParser()
             config.optionxform = str  # Preserve case
             config.read(self._ctx.config_file)
-            for section in config.sections():
-                for k, v in config.items(section):
-                    # Skip if the key exists in any section
-                    if self.get_var(k, False):
-                        continue
-                    # Account for empty section
-                    if not self.env.get(section, False):
-                        self.env[section] = {}
-                    self.env[section][k] = v
+            for k, v in config.items("config"):
+                if not self.get(k, None) and v:
+                    self[k.upper()] = v
         except Exception as e:
             utils.handle_exception(
                 e,
@@ -295,10 +267,8 @@ class EnvironmentVariables:
             )
 
     def _parse_library_env(self):
-        """Parses the Minitrino library's root `minitrino.env` file. All config from
-        this file is added to the 'MODULES' section of the environment
-        dictionary since this file explicitly defines the versions of the module
-        services."""
+        """Parses the Minitrino library's `minitrino.env` file. Lowest
+        precedence in environment variable order, loaded last."""
 
         env_file = os.path.join(self._ctx.minitrino_lib_dir, "minitrino.env")
         if not os.path.isfile(env_file):
@@ -308,76 +278,20 @@ class EnvironmentVariables:
                 f"present in that library?",
             )
 
-        # Check if modules section was added from Minitrino config file parsing
-        section = self.env.get("MODULES", None)
-        if not isinstance(section, dict):
-            self.env["MODULES"] = {}
-
         with open(env_file, "r") as f:
             for env_var in f:
                 env_var = utils.parse_key_value_pair(env_var, err_type=err.UserError)
                 if env_var is None:
                     continue
-                # Skip if the key exists in any section
-                if self.get_var(env_var[0], False):
-                    continue
-                self.env["MODULES"][env_var[0]] = env_var[1]
-
-    def _parse_user_env(self):
-        """Parses user-provided environment variables for the current
-        command."""
-
-        if not self._ctx._user_env:
-            return
-
-        user_env_dict = {}
-        for env_var in self._ctx._user_env:
-            env_var = utils.parse_key_value_pair(env_var, err_type=err.UserError)
-            user_env_dict[env_var[0]] = env_var[1]
-
-        # Loop through user-provided environment variables and check for matches
-        # in each section dict. If the variable key is in the section dict, it
-        # needs to be identified and replaced with the user's `--env` value,
-        # effectively overriding the original value.
-        #
-        # We build a new dict to prevent issues with changing dict size while
-        # iterating.
-        #
-        # Any variable keys that do not match with an existing key will be added
-        # to the section dict "EXTRA".
-
-        new_dict = {}
-        for section_k, section_v in self.env.items():
-            if not isinstance(section_v, dict):
-                raise err.MinitrinoError(
-                    f"Invalid environment dictionary. Expected nested dictionaries. "
-                    f"Received dictionary:\n"
-                    f"{json.dumps(self.env, indent=2)}"
-                )
-            delete_keys = []
-            for user_k, user_v in user_env_dict.items():
-                if user_k in section_v:
-                    if new_dict.get(section_k, None) is None:
-                        new_dict[section_k] = section_v
-                    new_dict[section_k][user_k] = user_v
-                    delete_keys.append(user_k)
-                else:
-                    new_dict[section_k] = section_v
-            for delete_key in delete_keys:
-                del user_env_dict[delete_key]
-
-        if user_env_dict:
-            self.env["EXTRA"] = {}
-            for k, v in user_env_dict.items():
-                self.env["EXTRA"][k] = v
+                if not self.get(env_var[0], None):
+                    self[env_var[0]] = env_var[1]
 
     def _log_env_vars(self):
         """Logs environment variables."""
 
-        if self.env:
-            self._ctx.logger.log(
-                f"Registered environment variables:\n{json.dumps(self.env, indent=2)}",
-                level=self._ctx.logger.verbose,
+        if self:
+            self._ctx.logger.verbose(
+                f"Registered environment variables:\n{json.dumps(self, indent=2)}",
             )
 
 
@@ -392,13 +306,10 @@ class Modules:
     - `data`: A dictionary of module information.
 
     ### Public Methods
-    - `get_running_modules()`: Returns a dictionary with the same information as
-      the `modules` attribute, but includes Docker labels and container objects
-      tied to the module."""
+    - `get_running_modules()`: Returns a list of running modules."""
 
     @utils.exception_handler
     def __init__(self, ctx=None):
-
         if not ctx:
             raise utils.handle_missing_param(list(locals().keys()))
 
@@ -416,56 +327,41 @@ class Modules:
         )
 
         if not containers:
-            return {}
+            return []
 
-        # Remove Trino container since it isn't a module
-        for i, container in enumerate(containers):
-            if container.name == "trino":
-                del containers[i]
-
-        names = []
-        label_sets = []
-        for i, container in enumerate(containers):
+        modules = []
+        for container in containers:
             label_set = {}
+            ids = ["admin-", "catalog-", "security-"]
             for k, v in container.labels.items():
-                if "catalog-" in v:
-                    names.append(v.lower().strip().replace("catalog-", ""))
-                elif "security-" in v:
-                    names.append(v.lower().strip().replace("security-", ""))
-                else:
-                    continue
-                label_set[k] = v
+                for _id in ids:
+                    if "com.starburst.tests" in k and _id in v:
+                        modules.append(v.lower().strip().replace(_id, ""))
+                        label_set[k] = v
+            # All containers except the trino container must have
+            # module-specific labels. The trino container only has module labels
+            # if a module applies labels to it
             if not label_set and container.name != "trino":
                 raise err.UserError(
                     f"Missing Minitrino labels for container '{container.name}'.",
                     f"Check this module's 'docker-compose.yml' file and ensure you are "
                     f"following the documentation on labels.",
                 )
-            label_sets.append(label_set)
 
-        running = {}
-        for name, label_set, container in zip(names, label_sets, containers):
-            if not isinstance(self.data.get(name), dict):
+        modules = list(set(modules))
+        for module in modules:
+            if not isinstance(self.data.get(module), dict):
                 raise err.UserError(
-                    f"Module '{name}' is running, but it is not found "
+                    f"Module '{module}' is running, but it is not found "
                     f"in the library. Was it deleted, or are you pointing "
                     f"Minitrino to the wrong location?"
                 )
-            if not running.get(name, False):
-                running[name] = self.data[name]
-            if not running.get("labels", False):
-                running[name]["labels"] = label_set
-            if not running.get(name).get("containers", False):
-                running[name]["containers"] = [container]
-            else:
-                running[name]["containers"].append(container)
-
-        return running
+        return modules
 
     def _load_modules(self):
         """Loads module data during instantiation."""
 
-        self._ctx.logger.log("Loading modules...", level=self._ctx.logger.verbose)
+        self._ctx.logger.verbose("Loading modules...")
 
         modules_dir = os.path.join(self._ctx.minitrino_lib_dir, MODULE_ROOT)
         if not os.path.isdir(modules_dir):
@@ -474,8 +370,9 @@ class Modules:
                 f"Are you pointing to a compatible Minitrino library?"
             )
 
-        # Loop through both catalog and security modules
+        # Loop through all module types
         sections = [
+            os.path.join(modules_dir, MODULE_ADMIN),
             os.path.join(modules_dir, MODULE_CATALOG),
             os.path.join(modules_dir, MODULE_SECURITY),
         ]
@@ -485,10 +382,9 @@ class Modules:
                 module_dir = os.path.join(section_dir, _dir)
 
                 if not os.path.isdir(module_dir):
-                    self._ctx.logger.log(
+                    self._ctx.logger.verbose(
                         f"Skipping file (expected a directory, not a file) "
                         f"at path: {module_dir}",
-                        level=self._ctx.logger.verbose,
                     )
                     continue
 
@@ -527,10 +423,9 @@ class Modules:
                     with open(json_file) as f:
                         metadata = json.load(f)
                 else:
-                    self._ctx.logger.log(
+                    self._ctx.logger.verbose(
                         f"No JSON metadata file for module '{module_name}'. "
                         f"Will not load metadata for module.",
-                        level=self._ctx.logger.verbose,
                     )
                 for k, v in metadata.items():
                     self.data[module_name][k] = v
@@ -550,7 +445,6 @@ class CommandExecutor:
 
     @utils.exception_handler
     def __init__(self, ctx=None):
-
         if not ctx:
             raise utils.handle_missing_param(list(locals().keys()))
 
@@ -601,9 +495,8 @@ class CommandExecutor:
     def _execute_in_shell(self, command="", **kwargs):
         """Executes a command in the host shell."""
 
-        self._ctx.logger.log(
+        self._ctx.logger.verbose(
             f"Executing command in shell:\n{command}",
-            level=self._ctx.logger.verbose,
         )
 
         process = subprocess.Popen(
@@ -622,21 +515,18 @@ class CommandExecutor:
             # do is to run the string through the `_strip_ansi()` function.
 
             started_stream = False
+            output = ""
             while True:
                 output_line = process.stdout.readline()
                 if output_line == "" and process.poll() is not None:
                     break
                 output_line = self._strip_ansi(output_line)
                 if not started_stream:
-                    self._ctx.logger.log(
-                        "Command Output:", level=self._ctx.logger.verbose
-                    )
+                    self._ctx.logger.verbose("Command Output:")
                     started_stream = True
-                self._ctx.logger.log(
-                    output_line, level=self._ctx.logger.verbose, stream=True
-                )
+                self._ctx.logger.verbose(output_line, stream=True)
+                output += output_line
 
-        output, _ = process.communicate()  # Get full output (stdout + stderr)
         if process.returncode != 0 and kwargs.get("trigger_error", True):
             raise err.MinitrinoError(
                 f"Failed to execute shell command:\n{command}\n"
@@ -660,10 +550,10 @@ class CommandExecutor:
                 f"container, but a container object was not provided."
             )
 
-        self._ctx.logger.log(
-            f"Executing command in container '{container.name}':\n{command}",
-            level=self._ctx.logger.verbose,
-        )
+        if not kwargs.get("suppress_output"):
+            self._ctx.logger.verbose(
+                f"Executing command in container '{container.name}':\n{command}",
+            )
 
         # Create exec handler and execute the command
         exec_handler = self._ctx.api_client.exec_create(
@@ -695,13 +585,9 @@ class CommandExecutor:
                 full_line += chunk[0]
                 if not kwargs.get("suppress_output", False):
                     if not started_stream:
-                        self._ctx.logger.log(
-                            "Command Output:", level=self._ctx.logger.verbose
-                        )
+                        self._ctx.logger.verbose("Command Output:")
                         started_stream = True
-                    self._ctx.logger.log(
-                        full_line, level=self._ctx.logger.verbose, stream=True
-                    )
+                    self._ctx.logger.verbose(full_line, stream=True)
                     full_line = ""
                 if chunk[1]:
                     full_line = chunk[1]
@@ -710,12 +596,19 @@ class CommandExecutor:
 
         # Catch lingering full line post-loop
         if not kwargs.get("suppress_output", False) and full_line:
-            self._ctx.logger.log(full_line, level=self._ctx.logger.verbose, stream=True)
+            self._ctx.logger.verbose(full_line, stream=True)
 
         # Get the exit code
         return_code = self._ctx.api_client.exec_inspect(exec_handler["Id"]).get(
             "ExitCode"
         )
+        # https://www.gnu.org/software/bash/manual/html_node/Exit-Status.html
+        if return_code == 126:
+            self._ctx.logger.warn(
+                f"The command exited with a 126 code which typically means an "
+                f"executable is not accessible or installed. Does this image have "
+                f"all required dependencies installed?\nCommand: {command}",
+            )
 
         if return_code != 0 and kwargs.get("trigger_error", True):
             raise err.MinitrinoError(
